#include <string.h>
#include "MEM.h"
#include "DBG.h"
#include "diksamc.h"

static CompilerList *st_compiler_list = NULL;

DKC_Compiler *
DKC_create_compiler(void)
{
    MEM_Storage storage;
    DKC_Compiler *compiler;

    storage = MEM_open_storage(0);
    compiler = MEM_storage_malloc(storage,
                                  sizeof(struct DKC_Compiler_tag));
    compiler->compile_storage = storage;
    compiler->package_name = NULL;
    compiler->require_list = NULL;
    compiler->rename_list = NULL;
    compiler->function_list = NULL;
    compiler->dvm_function_count = 0;
    compiler->dvm_function = NULL;
    compiler->dvm_class_count = 0;
    compiler->dvm_class = NULL;
    compiler->declaration_list = NULL;
    compiler->statement_list = NULL;
    compiler->class_definition_list = NULL;
    compiler->current_block = NULL;
    compiler->current_line_number = 1;
    compiler->current_class_definition = NULL;
    compiler->current_catch_clause = NULL;
    compiler->input_mode = DKC_FILE_INPUT_MODE;
    compiler->required_list = NULL;

    return compiler;
}

static CompilerList *
add_compiler_to_list(CompilerList *list, DKC_Compiler *compiler)
{
    CompilerList *pos;
    CompilerList *new_item;

    new_item = MEM_malloc(sizeof(CompilerList));
    new_item->compiler = compiler;
    new_item->next = NULL;

    if (list == NULL) {
        return new_item;
    }
    for (pos = list; pos->next; pos = pos->next)
        ;
    pos->next = new_item;

    return list;
}

static DKC_Compiler *
search_compiler(PackageName *package_name)
{
    CompilerList *pos;

    for (pos = st_compiler_list; pos; pos = pos->next) {
        if (dkc_compare_package_name(pos->compiler->package_name,
                                     package_name))
            break;
    }

    if (pos) {
        return pos->compiler;
    } else {
        return NULL;
    }
}

static void
make_search_path(int line_number, PackageName *package_name, char *buf)
{
    PackageName *pos;
    int len = 0;
    int prev_len = 0;
    int suffix_len;

    suffix_len = strlen(DIKSAM_REQUIRE_SUFFIX);
    buf[0] = '\0';
    for (pos = package_name; pos; pos = pos->next) {
        prev_len = len;
        len += strlen(pos->name);
        if (len > FILENAME_MAX - (2 + suffix_len)) {
            dkc_compile_error(line_number,
                              PACKAGE_NAME_TOO_LONG_ERR,
                              MESSAGE_ARGUMENT_END);
        }
        strcpy(&buf[prev_len], pos->name);
        if (pos->next) {
            buf[strlen(buf)] = FILE_SEPARATOR;
            len++;
        }
    }
    strcpy(&buf[len], DIKSAM_REQUIRE_SUFFIX);
}

static FILE*
get_require_fp(RequireList *req, char *found_path)
{
    char *search_path;
    char search_file[FILENAME_MAX];
    FILE *fp;
    SearchFileStatus status;
        
    search_path = getenv("DKM_REQUIRE_SEARCH_PATH");
    if (search_path == NULL) {
        search_path = ".";
    }
    make_search_path(req->line_number, req->package_name, search_file);

    status = dvm_search_file(search_path, search_file, found_path, &fp);
    
    if (status != SEARCH_FILE_SUCCESS) {
        if (status == SEARCH_FILE_NOT_FOUND) {
            dkc_compile_error(req->line_number,
                              REQUIRE_FILE_NOT_FOUND_ERR,
                              STRING_MESSAGE_ARGUMENT, "file", search_file,
                              MESSAGE_ARGUMENT_END);
        } else {
            dkc_compile_error(req->line_number,
                              REQUIRE_FILE_ERR,
                              INT_MESSAGE_ARGUMENT, "status", (int)status,
                              MESSAGE_ARGUMENT_END);
        }
    }

    return fp;
}

static DVM_Boolean
add_exe_to_pool(DVM_Executable *exe, DVM_ExecutablePool *pool)
{
    DVM_ExecutableList *new_item;
    DVM_ExecutableList *pos;
    DVM_ExecutableList *tail;

    for (pos = pool->list; pos; pos = pos->next) {
        if (dvm_compare_package_name(pos->executable->package_name,
                                     exe->package_name)
            && pos->executable->is_required == exe->is_required) {
            return DVM_FALSE;
        }
        tail = pos;
    }

    new_item = MEM_malloc(sizeof(DVM_ExecutableList));
    new_item->executable = exe;
    new_item->next = NULL;

    if (pool->list == NULL) {
        pool->list = new_item;
    } else {
        tail->next = new_item;
    }

    return DVM_TRUE;
}

static void
set_path_to_compiler(DKC_Compiler *compiler, char *path)
{
    compiler->path = MEM_storage_malloc(compiler->compile_storage,
                                        strlen(path) + 1);
    strcpy(compiler->path, path);
}

static DVM_Executable *
do_compile(DKC_Compiler *compiler, DVM_ExecutablePool *pool,
           char *path, DVM_Boolean is_required)
{
    extern FILE *yyin;
    extern int yyparse(void);
    RequireList *req_pos;
    DKC_Compiler *req_comp;
    DVM_Executable *exe;
    DVM_Executable *req_exe;
    FILE *fp;
    char found_path[FILENAME_MAX];
    DKC_Compiler *compiler_backup;

    compiler_backup = dkc_get_current_compiler();
    dkc_set_current_compiler(compiler);
    if (yyparse()) {
        fprintf(stderr, "Error ! Error ! Error !\n");
        exit(1);
    }

    for (req_pos = compiler->require_list; req_pos;
         req_pos = req_pos->next) {
        req_comp = search_compiler(req_pos->package_name);
        if (req_comp) {
            compiler->required_list
                = add_compiler_to_list(compiler->required_list, req_comp);
            continue;
        }
        req_comp = DKC_create_compiler();
        
        /* BUGBUG req_comp references parent compiler's MEM_storage */
        req_comp->package_name = req_pos->package_name;

        compiler->required_list
            = add_compiler_to_list(compiler->required_list, req_comp);
        st_compiler_list = add_compiler_to_list(st_compiler_list, req_comp);

        fp = get_require_fp(req_pos, found_path);
        set_path_to_compiler(req_comp, found_path);
        yyin = fp;
        
        req_exe = do_compile(req_comp, pool, found_path, DVM_TRUE);
    }

    dkc_fix_tree(compiler);
    exe = dkc_generate(compiler);
    if (path) {
        exe->path = MEM_strdup(path);
    } else {
        exe->path = NULL;
    }

    exe->is_required = is_required;
    if (!add_exe_to_pool(exe, pool)) {
        dvm_dispose_executable(exe);
    } else {
        /*
        dvm_disassemble(exe);
        */
    }

    dkc_set_current_compiler(compiler_backup);

    return exe;
}

static void
dispose_compiler_list(void)
{
    CompilerList *temp;

    while (st_compiler_list) {
        temp = st_compiler_list;
        st_compiler_list = temp->next;
        MEM_free(temp);
    }
}


DVM_ExecutablePool *
DKC_compile(DKC_Compiler *compiler, FILE *fp, char *path)
{
    extern FILE *yyin;
    DVM_ExecutablePool *pool;
    DVM_Executable *exe;

    DBG_assert(st_compiler_list == NULL,
               ("st_compiler_list != NULL(%p)", st_compiler_list));
    set_path_to_compiler(compiler, path);
    compiler->current_line_number = 1;
    compiler->input_mode = DKC_FILE_INPUT_MODE;

    yyin = fp;

    pool = MEM_malloc(sizeof(DVM_ExecutablePool));
    pool->list = NULL;
    
    exe = do_compile(compiler, pool, NULL, DVM_FALSE);
    exe->path = MEM_strdup(path);
    pool->top_level = exe;

    dispose_compiler_list();
    dkc_reset_string_literal_buffer();

    return pool;
}

PackageName *
create_one_package_name(DKC_Compiler *compiler,
                        char *str, int start_idx, int to_idx)
{
    PackageName *pn;
    int i;

    MEM_Storage storage = compiler->compile_storage;
    pn = MEM_storage_malloc(storage, sizeof(PackageName));
    pn->name = MEM_storage_malloc(storage, to_idx - start_idx + 1);

    for (i = 0; i < to_idx - start_idx; i++) {
        pn->name[i] = str[start_idx + i];
    }
    pn->name[i] = '\0';
    pn->next = NULL;

    return pn;
}

static PackageName *
string_to_package_name(DKC_Compiler *compiler, char *str)
{
    int start_idx;
    int i;
    PackageName *pn;
    PackageName *top;
    PackageName *tail = NULL;

    for (start_idx = i = 0; str[i] != '\0'; i++) {
        if (str[i] == '.') {
            pn = create_one_package_name(compiler, str, start_idx, i);
            start_idx = i + 1;
            if (tail) {
                tail->next = pn;
            } else {
                top = pn;
            }
            tail = pn;
        }
    }
    pn = create_one_package_name(compiler, str, start_idx, i);
    if (tail) {
        tail->next = pn;
    } else {
        top = pn;
    }

    return top;
}

DVM_ExecutableList *
DKC_dynamic_compile(DKC_Compiler *compiler, FILE *fp,
                    char *package_name, DVM_ExecutablePool *pool,
                    char *path)
{
    extern FILE *yyin;
    DVM_ExecutableList *tail;
    DVM_Executable *exe;

    DBG_assert(st_compiler_list == NULL,
               ("st_compiler_list != NULL(%p)", st_compiler_list));

    for (tail = pool->list; tail->next; tail = tail->next)
        ;

    compiler->package_name = string_to_package_name(compiler, package_name);
    set_path_to_compiler(compiler, path);
    compiler->current_line_number = 1;
    compiler->input_mode = DKC_FILE_INPUT_MODE;

    yyin = fp;

    exe = do_compile(compiler, pool, path, DVM_FALSE);

    dispose_compiler_list();
    dkc_reset_string_literal_buffer();

    return tail->next;
}


DVM_Executable *
DKC_compile_string(DKC_Compiler *compiler, char **lines,
                   DVM_Boolean is_required)
{
    extern int yyparse(void);
    DVM_ExecutablePool *pool;
    DVM_Executable *exe;

    dkc_set_source_string(lines);
    compiler->current_line_number = 1;
    compiler->input_mode = DKC_STRING_INPUT_MODE;

    pool = MEM_malloc(sizeof(DVM_ExecutablePool));
    pool->list = NULL;
    exe = do_compile(compiler, pool, NULL, is_required);

    dkc_reset_string_literal_buffer();

    return exe;
}

static void
dispose_compiler_sub(DKC_Compiler *compiler)
{
    FunctionDefinition *fd_pos;
    CompilerList *temp;

    for (fd_pos = compiler->function_list; fd_pos; fd_pos = fd_pos->next) {
        MEM_free(fd_pos->local_variable);
    }

    while (compiler->required_list) {
        dispose_compiler_sub(compiler->required_list->compiler);
        temp = compiler->required_list;
        compiler->required_list = temp->next;
        MEM_free(temp);
    }

    MEM_dispose_storage(compiler->compile_storage);
}

void
DKC_dispose_compiler(DKC_Compiler *compiler)
{
    dispose_compiler_sub(compiler);
}
